#include "fw/dynmx/servo.h"

using namespace schpin;

std::optional< servo_master_message >
dynmx_servo::on_servo_receive( clk_time, s_ctl_id sid, const dynmx_buffer& buff )
{

        context_.on_receive( sid, buff );

        return {};
}

void dynmx_servo::tick( clk_time now, servo_visitor& vis )
{
        task_ << match(
            [&]( idle_task ) {},
            [&]( activate_task at ) {
                    at.i = context_.full_poll( at.i, vis );
                    if ( at.i == 0 ) {
                            task_ = idle_task{};
                    } else {
                            task_ = at;
                    }
            },
            [&]( goto_task gt ) {
                    clk_time t_time = std::clamp( now, gt.from_time, gt.to_time );
                    float    mapped = map_val< float >(
                        t_time, gt.from_time, gt.to_time, *gt.from_pos, *gt.to_pos );

                    context_.set_item< DYNMX_GOAL_POS >( *dynmx_angle::make( mapped ), vis );

                    SCHPIN_INFO_LOG( "dynmx", "goto: " << mapped );

                    if ( now < gt.to_time ) {
                            return;
                    }

                    if ( abs( static_cast< int >( *context_.get_item< DYNMX_PRESENT_POS >() ) -
                              static_cast< int >( *gt.to_pos ) ) > 15 ) {
                            return;
                    }

                    task_ = idle_task{};
            },
            [&]( spin_task ) {} );
}

void dynmx_servo::on_event(
    ctl_id,
    clk_time                   now,
    dynmx_servo_events_variant var,
    servo_visitor&             vis )
{
        var << match(
            [&]( dynmx_goto_event e ) {
                    context_.set_item< DYNMX_TORQUE_EN >( dynmx_false, vis );
                    context_.set_item< DYNMX_OPER_MODE >( dynmx_position_control_mode, vis );
                    context_.set_item< DYNMX_TORQUE_EN >( dynmx_true, vis );

                    task_ = goto_task{
                        context_.get_item< DYNMX_PRESENT_POS >(),
                        e.goal_pos,
                        now,
                        now + clk_time{ *e.t } };
            },
            [&]( dynmx_spin_event ) {},
            [&]( dynmx_config_event ) {} );
}

void dynmx_servo::poll( clk_time, std::size_t i_offset, servo_visitor& vis )
{
        poll_i_       = ( poll_i_ + 1 ) % 13;
        std::size_t i = ( i_offset + poll_i_ ) % 13;
        switch ( i ) {
                case 4:
                        context_.query_item< DYNMX_PRESENT_VOLTAGE >( vis );
                        break;
                case 8:
                        context_.query_item< DYNMX_PRESENT_TEMP >( vis );
                        break;
                default:
                        context_.query_item< DYNMX_PRESENT_POS >( vis );
                        break;
        }
}
em::either< dynmx_servo_events_variant, servo_master_message >
dynmx_servo::on_servo_packet( ctl_id cid, master_servo_command_variant var )
{
        using return_type = em::either< dynmx_servo_events_variant, servo_master_message >;

        return var << tuple_match(
                   [&]( cmd_tag< GOTO >, bangle a, btime t ) -> return_type {
                           return handle_in_to_se( converter_, cid, GOTO, a )
                               .convert_left( [&]( dynmx_angle an ) -> dynmx_servo_events_variant {
                                       return dynmx_goto_event{ an, t };
                               } );
                   },
                   [&]( cmd_tag< SPIN >, bangular_velocity a, btime t ) -> return_type {
                           return handle_in_to_se( converter_, cid, SPIN, a )
                               .convert_left(
                                   [&]( dynmx_angular_velocity av ) -> dynmx_servo_events_variant {
                                           return dynmx_spin_event{ av, t };
                                   } );
                   },
                   [&]( cmd_tag< POSITION > ) -> return_type {
                           return make_servo_message< POSITION >(
                               converter_.se_to_in( context_.get_item< DYNMX_PRESENT_POS >() ) );
                   },
                   [&]( cmd_tag< TEMP > ) -> return_type {
                           return make_servo_message< TEMP >(
                               converter_.se_to_in( context_.get_item< DYNMX_PRESENT_TEMP >() ) );
                   },
                   [&]( cmd_tag< VOLTAGE > ) -> return_type {
                           return make_servo_message< VOLTAGE >( converter_.se_to_in(
                               context_.get_item< DYNMX_PRESENT_VOLTAGE >() ) );
                   },
                   [&]( cmd_tag< VELOCITY > ) -> return_type {
                           return make_servo_message< VELOCITY >(
                               converter_.se_to_in( context_.get_item< DYNMX_PRESENT_VEL >() ) );
                   },
                   [&]( cmd_tag< TASK > ) -> return_type {
                           servo_task_enum e = task_ << match(
                                                   [&]( idle_task ) {
                                                           return TASK_IDLE;
                                                   },
                                                   [&]( activate_task ) {
                                                           return TASK_ACTIVATE;
                                                   },
                                                   [&]( goto_task ) {
                                                           return TASK_GOTO;
                                                   },
                                                   [&]( spin_task ) {
                                                           return TASK_SPIN;
                                                   } );

                           return make_servo_message< TASK >( e );
                   },
                   [&]( cmd_tag< CONFIG_GET >, config_addr addr ) -> return_type {
                           return make_servo_message< CONFIG_GET >(
                               addr,
                               config_map_.get_buffered_val< config_buffer >(
                                   static_cast< dynmx_config_enum >( addr ) ) );
                   },
                   [&]( cmd_tag< CONFIG_SET >,
                        config_addr   addr,
                        config_buffer buffr ) -> return_type {
                           return dynmx_servo_events_variant{ dynmx_config_event{
                               static_cast< dynmx_config_enum >( addr ), buffr } };
                   } );
}

void dynmx_servo::dump_info()
{
        context_.dump_info();
}
