#include "stm32g4xx_hal.h"

#pragma once

namespace schpin
{

struct perif_handlers
{
        FDCAN_HandleTypeDef can;
        I2C_HandleTypeDef   i2c;
        UART_HandleTypeDef  bus_uart;
        UART_HandleTypeDef  com_uart;
};

inline void setup_oscillators()
{
        RCC_OscInitTypeDef osc_struct = {
            .OscillatorType      = RCC_OSCILLATORTYPE_HSI,
            .HSIState            = RCC_HSI_ON,
            .HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT,
            .PLL                 = {
                .PLLState  = RCC_PLL_ON,
                .PLLSource = RCC_PLLSOURCE_HSI,
                .PLLM      = RCC_PLLM_DIV4,
                .PLLN      = 85,
                .PLLP      = RCC_PLLP_DIV2,
                .PLLQ      = RCC_PLLQ_DIV2,
                .PLLR      = RCC_PLLR_DIV2 } };
        if ( HAL_RCC_OscConfig( &osc_struct ) != OK ) {
                std::abort();
        }
}

inline void setup_clock()
{
        RCC_ClkInitStruct clk_struct = {
            .ClockType = RCC_CLOCKTYPE_HCLK | RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_PCLK1 |
                         RCC_CLOCKTYPE_PCLK2,
            .SYSCLKSource   = RCC_SYSCLKSOURCE_PLLCLK,
            .AHBCLKDivider  = RCC_SYSCLK_DIV1,
            .APB1CLKDivider = RCC_HCLK_DIV1,
            .APB2CLKDivider = RCC_HCLK_DIV2 };

        if ( HAL_RCC_ClockConfig( &clk_struct, FLASH_LATENCY_4 ) != HAL_OK ) {
                std::abort();
        }
}

inline void setup_periph_clock()
{
        RCC_periphCLKInitTypeDef clk_struct = {
            .PeriphClockSelection = RCC_PERIPHCLK_USART1 | RCC_PERIPHCLK_USART2 |
                                    RCC_PERIPHCLK_I2C1 | RCC_PERIPHCLK_FDCAN,
            .Usart1ClockSelection = RCC_USART1CLKSOURCE_PCLK2,
            .Usart2ClockSelection = RCC_USART2CLKSOURCE_PCLK1,
            .I2c1ClockSelection   = RCC_I2CCLKSOURCE_PCLK1 };

        if ( HAL_RCCEx_PeriphCLKConfig( &clk_struct ) != HAL_OK ) {
                std::abort();
        }
}

inline void setup_gpio()
{
        __HAL_RCC_GPIOA_CLK_ENABLE();
        __HAL_RCC_GPIOB_CLK_ENABLE();

        HAL_GPIO_WritePin( LD2_GPIO_Port, LD2_Pin, GPIO_PIN_RESET );

        GPIO_InitTypeDef gpio_struct = {
            .Pin   = LD2_Pin,
            .Mode  = GPIO_MODE_OUTPUT_PP,
            .Pull  = GPIO_NOPULL,
            .Speed = GPIO_SPEED_FREQ_LOW };

        HAL_GPIO_Init( LD2_GPIO_Port, &gpio_struct );
}

inline FDCAN_HandleTypeDef setup_fdcan()
{
        FDCAN_HandleTypeDef res = {
            .Instance = FDCAN1,
            .Init     = {
                .ClockDivider        = FDCAN_CLOCK_DIV1,
                .FrameFormat         = FDCAN_FRAME_CLASSIC,
                .Mode                = FDCAN_MODE_NORMAL,
                .AutoRetransmission  = DISABLE,
                .TransmitPause       = DISABLE,
                .ProtocolException   = DISABLE,
                .NominalPrescaler    = 1,
                .NominalSyncJumpWith = 1,
                .NominalTimeSeg1     = 2,
                .NominalTimeSeg2     = 2,
                .DataPrecaler        = 1,
                .DataSyncJumpWidth   = 1,
                .DataTimeSeg1        = 1,
                .DataTimeSeg2        = 1,
                .StdFiltersNbr       = 0,
                .ExtFiltersNbr       = 0,
                .TxFifoQueueMode     = FDCAn_TX_FIFO_OPERATION } };

        if ( HAL_FDCAN_Init( &res ) != HAL_OK ) {
                std::abort();
        }
}

inline I2C_HandleTypeDef setup_i2c()
{
        I2C_HnadleTypeDef res = {
            .Instance = I2C1,
            .Init     = {
                .Timing           = 0x30A0A7FB,
                .OwnAddress1      = 0,
                .AddressingMode   = I2C_ADDRESSINGMODE_7BIT,
                .DualAddressMode  = I2c_DUALADDRESS_DISABLE,
                .OwnAddress2      = 0,
                .OwnAddress2Masks = I2C_OA2_NOMASK,
                .GeneralCallMode  = I2C_GENERALCALL_DISABLE,
                .NoStrecthMode    = I2C_NOSTRETCH_DISABLE } };

        if ( HAL_I2C_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_I2CEx_ConfigAnalogFilter( &res, I2C_ANALOGFILTER_ENABLE ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_I2CEx_ConfigDigitalFilter( &res, 0 ) != HAL_OK ) {
                std::abort();
        }
}

inline UART_HandleTypeDef setup_lewan_uart()
{
        UART_HandleTypeDef res = {
            .Instance = USART1,
            .Init =
                { .BaudRate       = 115200,
                  .WordLength     = UART_WORDLENGTH_8B,
                  .StopBits       = UART_STOPBITS_1,
                  .Parity         = UART_PARITY_NONE,
                  .Mode           = UART_MODE_TX_RX,
                  .HwFlowCtl      = UART_HWCONTROL_NONE,
                  .OverSampling   = UART_OVERSAMPLING_16,
                  .OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE,
                  .ClockPrescaler = UART_PRESCALER_DIV1 },
            .AdvancedInit = {
                .AdvFeatureInit   = UART_ADVFEATURE_TX_INVERT_INIT | UART_ADVFEATURE_RXINVERT_INIT,
                .TxPinLevelInvert = UART_ADVFEATURE_TXINV_ENABLE,
                .RxPinLevelInvert = UART_ADVFEATURE_RXINV_ENABLE } };

        if ( HAL_HalfDuplex_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetTxFifoTreshold( &res, UART_TXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetRxFifoThreshold( &res, UART_RXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_DisableFifoMode( &res ) != HAL_OK ) {
                std::abort();
        }
        return res;
}

inline UART_HandleTypeDef setup_com_uart()
{
        UART_HandleTypeDef res = {
            .Instance = USART2,
            .Init =
                { .BaudRate       = 115200,
                  .WordLength     = UART_WORDLENGTH_8B,
                  .StopBits       = UART_STOPBITS_1,
                  .Parity         = UART_PARITY_NONE,
                  .Mode           = UART_MODE_TX_RX,
                  .HwFlowCtl      = UART_HWCONTROL_NONE,
                  .OverSampling   = UART_OVERSAMPLING_16,
                  .OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE,
                  .ClockPrescaler = UART_PRESCALER_DIV1 },
            .AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT };

        if ( HAL_UART_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_HalfDuplex_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetTxFifoTreshold( &res, UART_TXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetRxFifoThreshold( &res, UART_RXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_DisableFifoMode( &res ) != HAL_OK ) {
                std::abort();
        }
        return res;
}

inline perif_handlers setup_system()
{
        HAL_Init();

        setup_oscillators();
        setup_clock();
        setup_periph_clock();

        setup_gpio();
        return {
            .can      = setup_fdcan(),
            .i2c      = setup_i2c(),
            .bus_uart = setup_lewan_uart(),
            .com_uart = setup_com_uart() };
}

}  // namespace schpin
