// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "test/servo_runner.h"

#include <rclcpp/rclcpp.hpp>

using namespace schpin;

// TODO:
//
//    make extension to protocol that allows json -> protocol serialization

// TODO:
//    make sleep function for tests
//    properly handle errors in run_file, now it loops infiniitely

em::either< servo_id, runner_error > find_dynmx_servo( serial_connection ser )
{
        dynmx_serial dser( ser );
        // 254 is broadcast adress
        auto                            id = dynmx_servo_id::get< 254 >();
        std::optional< dynmx_servo_id > opt_res;

        dser.send( make_master_dynmx_msg< DYNMX_PING >( id ) );

        dynmx_master_message reply = dser.read_reply();

        dynmx_master_protocol::extract( reply ).match(
            em::uncurry |
                [&]( dynmx_header, dynmx_servo_id sid, auto, dynmx_crc ) {
                        opt_res = sid;
                },
            [&]( auto err ) {
                    SCHPIN_ERROR_LOG( "servo_runner", err << ", msg: " << reply );
            } );

        if ( !opt_res ) {
                return E( runner_error ) << "Failed to find id";
        }
        return servo_id{ **opt_res };
}

em::either< servo_id, runner_error > find_servo( servo_type stype, serial_connection ser )
{
        switch ( stype ) {
                case DYNMX:
                        return find_dynmx_servo( ser );
                default:
                        return E( runner_error )
                               << "AUto find for this type of servo not implemented";
        }
}

int main( int argc, char* argv[] )
{

        rclcpp::init( argc, argv );

        auto node = rclcpp::Node::make_shared(
            "servo_runner",
            rclcpp::NodeOptions()
                .enable_rosout( true )
                .allow_undeclared_parameters( true )
                .automatically_declare_parameters_from_overrides( true ) );

        using tmp_type = em::either<
            std::tuple< servo_tester_config, std::optional< serial_connection > >,
            runner_error >;

        get_config< servo_tester_config >( *node, "" )
            .bind_left( [&]( const servo_tester_config& conf ) -> tmp_type {
                    if ( conf.opt_dev ) {
                            return serial_connection::make( *conf.opt_dev )
                                .convert_left( [&]( serial_connection c ) {
                                        return std::make_tuple( conf, std::make_optional( c ) );
                                } )
                                .construct_right< runner_error >();
                    }
                    return std::make_tuple( conf, std::optional< serial_connection >() );
            } )
            .bind_left(
                em::uncurry | [&]( servo_tester_config                con,
                                   std::optional< serial_connection > opt_ser ) -> tmp_type {
                        if ( opt_ser ) {
                                return find_servo( con.type, *opt_ser )
                                    .convert_left( [&]( servo_id sid ) {
                                            con.opt_id = sid;
                                            return std::make_tuple( con, opt_ser );
                                    } );
                        }
                        con.opt_id = 0;
                        return std::make_tuple( con, opt_ser );
                } )
            .match(
                em::uncurry |
                    [&]( servo_tester_config conf, std::optional< serial_connection > opt_ser ) {
                            servo_runner runner{ conf, opt_ser };

                            if ( conf.opt_source ) {
                                    runner.run_file( *conf.opt_source );
                                    return;
                            }

                            while ( true ) {
                                    runner.tick();
                            }
                    },
                [&]( auto e ) {
                        std::cerr << e;
                } );
}
